==> kite-1.0b7.tar.gz: kite-1.0b7/modules/__internal/_cinvoke_function.c <==
/*****************************************************************************
 * Copyright (c) 2008, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "kite_object.h"
#include "vm/kite_vm.h"
#include "config.h"

#ifdef BUILD_CINVOKE
#include "cinvoke/cinvoke.h"

/*****************************************************************************
 * Class constructor
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeFunction_construct) {
    kite_object_t *ctxObj, *retString, *paramString;
    CInvContext *ctx;
    CInvFunction *func;
    char *ret, *params;
    kite_object_t *typeList;
    
    KITE_GET_METHOD_ARGUMENT(ctxObj, 1);
    KITE_GET_METHOD_ARGUMENT(retString, 2);
    KITE_GET_METHOD_ARGUMENT(paramString, 3);
    
    ctx = (CInvContext*)ctxObj->builtin_data.filevalue;
    ret = KITE_GET_STRING_VALUE(retString);
    params = KITE_GET_STRING_VALUE(paramString);

    func = cinv_function_create(ctx, CINV_CC_CDECL, ret, params);
    typeList = kite_new_list(thd);
    kite_append_list(thd, typeList, retString);
    kite_append_list(thd, typeList, paramString);
    
    this->builtin_data.listvalue.car = 
        (kite_object_t*)ctx;
    this->builtin_data.listvalue.cdr =
        (kite_object_t*)func;
    this->builtin_data.listvalue.iterator = typeList;
}

/*****************************************************************************
 * Class destructor
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeFunction_destruct) {
    KITE_NO_ARGS;
    (void)thd;
    
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvFunction *func = (CInvFunction*)this->builtin_data.listvalue.cdr;
    
    cinv_function_delete(ctx, func);
    kite_dereference_object(this->builtin_data.listvalue.iterator);
}

/*****************************************************************************
 * Invoke function
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeFunction_invoke) {
    kite_object_t *entry, *params;
    char *ret_types_string, *param_type_string;
    void *entrypoint, **param_buf = NULL, *ret_buf = NULL;
    int size = 0;
    
    KITE_GET_METHOD_ARGUMENT(entry, 1);
    KITE_GET_METHOD_ARGUMENT(params, 2);
    
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvFunction *func = (CInvFunction*)this->builtin_data.listvalue.cdr;
    kite_object_t *param_types = this->builtin_data.listvalue.iterator, *tmp;
    entrypoint = (void*)KITE_GET_INTEGER(entry);
    
    args = param_types;
    KITE_GET_METHOD_ARGUMENT(tmp, 1);
    ret_types_string = KITE_GET_STRING_VALUE(tmp);
    KITE_GET_METHOD_ARGUMENT(tmp, 2);
    param_type_string = KITE_GET_STRING_VALUE(tmp);
    
    args = params;
    while(*param_type_string)
    {
        KITE_GET_METHOD_ARGUMENT(tmp, ++size);
        param_buf = realloc(param_buf, sizeof(void*)*size);
        switch(*param_type_string)
        {
            case 'c':
            {
                param_buf[size - 1] = malloc(sizeof(char));
                *(char*)(param_buf[size - 1]) = (char)KITE_GET_INTEGER(tmp);
                break;
            }
            case 's':
            {
                param_buf[size - 1] = malloc(sizeof(short));
                *(short*)(param_buf[size - 1]) = (short)KITE_GET_INTEGER(tmp);
                break;
            }
            case 'i':
            {
                param_buf[size - 1] = malloc(sizeof(int));
                *(int*)(param_buf[size - 1]) = (int)KITE_GET_INTEGER(tmp);
                break;
            }
            case 'l':
            {
                param_buf[size - 1] = malloc(sizeof(long));
                *(long*)(param_buf[size - 1]) = (long)KITE_GET_INTEGER(tmp);
                break;
            }
            case 'p':
            {
                param_buf[size - 1] = malloc(sizeof(long));
                *(long*)(param_buf[size - 1]) = (long)KITE_GET_INTEGER(tmp);
                break;
            }
            case 't':
            {
                param_buf[size - 1] = malloc(sizeof(long));
                *(char**)(param_buf[size - 1]) = KITE_GET_STRING_VALUE(tmp);
                break;
            }
            case '2':
            {
                param_buf[size - 1] = malloc(sizeof(cinv_int16_t));
                *(cinv_int16_t*)(param_buf[size - 1]) = (cinv_int16_t)KITE_GET_INTEGER(tmp);
                break;
            }
            case '4':
            {
                param_buf[size - 1] = malloc(sizeof(cinv_int32_t));
                *(cinv_int32_t*)(param_buf[size - 1]) = (cinv_int32_t)KITE_GET_INTEGER(tmp);
                break;
            }
            case '8':
            {
                param_buf[size - 1] = malloc(sizeof(cinv_int64_t));
                *(cinv_int64_t*)(param_buf[size - 1]) = (cinv_int64_t)KITE_GET_INTEGER(tmp);
                break;
            }
            case 'e':
            {
                param_buf[size - 1] = malloc(sizeof(long long));
                *(long long*)(param_buf[size - 1]) = (long long)KITE_GET_INTEGER(tmp);
                break;
            }
            case 'f':
            {
                param_buf[size - 1] = malloc(sizeof(float));
                *(float*)(param_buf[size - 1]) = (float)KITE_GET_FLOAT(tmp);
                break;
            }
            case 'd':
            {
                param_buf[size - 1] = malloc(sizeof(double));
                *(double*)(param_buf[size - 1]) = (double)KITE_GET_FLOAT(tmp);
                break;
            }
            default:
            {
                assert(0);
            }
        }
        param_type_string++;
    }
    
    int sz = 0;
    switch(*ret_types_string)
    {
        case 'c':
        {
            sz = sizeof(char);
            break;
        }
        case 's':
        {
            sz = sizeof(short);
            break;
        }
        case 'i':
        {
            sz = sizeof(int);
            break;
        }
        case 'l':
        {
            sz = sizeof(long);
            break;
        }
        case 'e':
        {
            sz = sizeof(long long);
            break;
        }
        case 'f':
        {
            sz = sizeof(float);
            break;
        }
        case 'd':
        {
            sz = sizeof(double);
            break;
        }
        case 'p':
        {
            sz = sizeof(void*);
            break;
        }
        case 't':
        {
            sz = sizeof(void*);
            break;
        }
        case '2':
        {
            sz = sizeof(cinv_int16_t);
            break;
        }
        case '4':
        {
            sz = sizeof(cinv_int32_t);
            break;
        }
        case '8':
        {
            sz = sizeof(cinv_int64_t);
            break;
        }
        case 0:
        {
            // void
            break;
        }
        default:
        {
            assert(0);
        }
    }

    if (sz) ret_buf = malloc(sz);
    cinv_function_invoke(ctx, func, entrypoint, ret_buf, param_buf);
    
    for(; size > 0; size--)
    {
        free(param_buf[size - 1]);
    }
    free(param_buf);
    
    if (ret_buf)
    {
        switch(*ret_types_string)
        {
            case 'c':
            {
                tmp = kite_new_integer(thd, *(char*)ret_buf);
                break;
            }
            case 's':
            {
                tmp = kite_new_integer(thd, *(short*)ret_buf);
                break;
            }
            case 'i':
            {
                tmp = kite_new_integer(thd, *(int*)ret_buf);
                break;
            }
            case 'l':
            {
                tmp = kite_new_integer(thd, *(long*)ret_buf);
                break;
            }
            case 'e':
            {
                tmp = kite_new_integer(thd, *(long long*)ret_buf);
                break;
            }
            case 'f':
            {
                tmp = kite_new_float(thd, *(float*)ret_buf);
                break;
            }
            case 'd':
            {
                tmp = kite_new_float(thd, *(double*)ret_buf);
                break;
            }
            case 'p':
            {
                tmp = kite_new_integer(thd, (long)*(void**)ret_buf);
                break;
            }
            case 't':
            {
                if (*(char**)ret_buf)
                    tmp = kite_new_string(thd, *(char**)ret_buf);
                else
                    tmp = kite_new_null(thd);
                break;
            }
            case '2':
            {
                tmp = kite_new_integer(thd, *(cinv_int16_t*)ret_buf);
                break;
            }
            case '4':
            {
                tmp = kite_new_integer(thd, *(cinv_int32_t*)ret_buf);
                break;
            }
            case '8':
            {
                tmp = kite_new_integer(thd, *(cinv_int64_t*)ret_buf);
                break;
            }
            default:
            {
                assert(0);
            }
        }

        free(ret_buf);
        kite_vm_push(thd, tmp);
    }
    else
    {
        kite_vm_push(thd, kite_new_null(thd));
    }
}
#endif

KITE_MODULE_INITIALIZER(_cinvoke_function) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "_cinvoke_function");
    kite_set_docstring(newclass, "CInvokeFunction internal code.");

#ifdef BUILD_CINVOKE    
    /* Add methods. */
    kite_add_method(thread, newclass, "__construct__",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeFunction_construct, 
                                  "Class constructor.", 3,
                                  "ctx", "CInvokeContext instance",
                                  "ret", "Return parameter format",
                                  "param", "Input parameter format"));
    kite_add_method(thread, newclass, "destruct",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeFunction_destruct, 
                                  "Free associated memory and resources.", 0));
    kite_add_method(thread, newclass, "invoke",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeFunction_invoke, 
                                  "Retrieves function pointer for given function.", 2,
                                  "entry", "Function entry point.",
                                  "params", "List of parameters."));
#endif

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "_cinvoke_function", 1, NULL);
    kite_set_property(thread, parent, "_cinvoke_function", newclass);
}

==> kite_1.0.2.orig.tar.gz: kite-1.0.2/modules/__internal/_cinvoke_function.c <==
/*****************************************************************************
 * Copyright (c) 2008, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "kite_object.h"
#include "vm/kite_vm.h"
#include "config.h"

#ifdef BUILD_CINVOKE
#include "cinvoke/cinvoke.h"

/*****************************************************************************
 * Class constructor
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeFunction_construct) {
    kite_object_t *ctxObj, *retString, *paramString;
    CInvContext *ctx;
    CInvFunction *func;
    char *ret, *params;
    kite_object_t *typeList;
    
    KITE_GET_METHOD_ARGUMENT(ctxObj, 1);
    KITE_GET_METHOD_ARGUMENT(retString, 2);
    KITE_GET_METHOD_ARGUMENT(paramString, 3);
    
    ctx = (CInvContext*)ctxObj->builtin_data.filevalue;
    ret = KITE_GET_STRING_VALUE(retString);
    params = KITE_GET_STRING_VALUE(paramString);

    func = cinv_function_create(ctx, CINV_CC_CDECL, ret, params);
    typeList = kite_new_list(thd);
    kite_append_list(thd, typeList, retString);
    kite_append_list(thd, typeList, paramString);
    
    this->builtin_data.listvalue.car = 
        (kite_object_t*)ctx;
    this->builtin_data.listvalue.cdr =
        (kite_object_t*)func;
    this->builtin_data.listvalue.iterator = typeList;
}

/*****************************************************************************
 * Class destructor
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeFunction_destruct) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvFunction *func = (CInvFunction*)this->builtin_data.listvalue.cdr;
    
    KITE_NO_ARGS;
    (void)thd;
    
    cinv_function_delete(ctx, func);
    kite_dereference_object(this->builtin_data.listvalue.iterator);
}

/*****************************************************************************
 * Invoke function
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeFunction_invoke) {
    kite_object_t *entry, *params;
    char *ret_types_string, *param_type_string;
    void *entrypoint, **param_buf = NULL, *ret_buf = NULL;
    int size = 0;
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvFunction *func = (CInvFunction*)this->builtin_data.listvalue.cdr;
    kite_object_t *param_types = this->builtin_data.listvalue.iterator, *tmp;
    int sz = 0;
    
    KITE_GET_METHOD_ARGUMENT(entry, 1);
    KITE_GET_METHOD_ARGUMENT(params, 2);
    
    entrypoint = (void*)KITE_GET_INTEGER(entry);
    
    args = param_types;
    KITE_GET_METHOD_ARGUMENT(tmp, 1);
    ret_types_string = KITE_GET_STRING_VALUE(tmp);
    KITE_GET_METHOD_ARGUMENT(tmp, 2);
    param_type_string = KITE_GET_STRING_VALUE(tmp);
    
    args = params;
    while(*param_type_string)
    {
        KITE_GET_METHOD_ARGUMENT(tmp, ++size);
        param_buf = realloc(param_buf, sizeof(void*)*size);
        switch(*param_type_string)
        {
            case 'c':
            {
                param_buf[size - 1] = malloc(sizeof(char));
                *(char*)(param_buf[size - 1]) = (char)KITE_GET_INTEGER(tmp);
                break;
            }
            case 's':
            {
                param_buf[size - 1] = malloc(sizeof(short));
                *(short*)(param_buf[size - 1]) = (short)KITE_GET_INTEGER(tmp);
                break;
            }
            case 'i':
            {
                param_buf[size - 1] = malloc(sizeof(int));
                *(int*)(param_buf[size - 1]) = (int)KITE_GET_INTEGER(tmp);
                break;
            }
            case 'l':
            {
                param_buf[size - 1] = malloc(sizeof(long));
                *(long*)(param_buf[size - 1]) = (long)KITE_GET_INTEGER(tmp);
                break;
            }
            case 'p':
            {
                param_buf[size - 1] = malloc(sizeof(long));
                *(long*)(param_buf[size - 1]) = (long)KITE_GET_INTEGER(tmp);
                break;
            }
            case 't':
            {
                param_buf[size - 1] = malloc(sizeof(long));
                *(char**)(param_buf[size - 1]) = KITE_GET_STRING_VALUE(tmp);
                break;
            }
            case '2':
            {
                param_buf[size - 1] = malloc(sizeof(cinv_int16_t));
                *(cinv_int16_t*)(param_buf[size - 1]) = (cinv_int16_t)KITE_GET_INTEGER(tmp);
                break;
            }
            case '4':
            {
                param_buf[size - 1] = malloc(sizeof(cinv_int32_t));
                *(cinv_int32_t*)(param_buf[size - 1]) = (cinv_int32_t)KITE_GET_INTEGER(tmp);
                break;
            }
            case '8':
            {
                param_buf[size - 1] = malloc(sizeof(cinv_int64_t));
                *(cinv_int64_t*)(param_buf[size - 1]) = (cinv_int64_t)KITE_GET_INTEGER(tmp);
                break;
            }
            case 'e':
            {
                param_buf[size - 1] = malloc(sizeof(long long));
                *(long long*)(param_buf[size - 1]) = (long long)KITE_GET_INTEGER(tmp);
                break;
            }
            case 'f':
            {
                param_buf[size - 1] = malloc(sizeof(float));
                *(float*)(param_buf[size - 1]) = (float)KITE_GET_FLOAT(tmp);
                break;
            }
            case 'd':
            {
                param_buf[size - 1] = malloc(sizeof(double));
                *(double*)(param_buf[size - 1]) = (double)KITE_GET_FLOAT(tmp);
                break;
            }
            default:
            {
                assert(0);
            }
        }
        param_type_string++;
    }
    
    switch(*ret_types_string)
    {
        case 'c':
        {
            sz = sizeof(char);
            break;
        }
        case 's':
        {
            sz = sizeof(short);
            break;
        }
        case 'i':
        {
            sz = sizeof(int);
            break;
        }
        case 'l':
        {
            sz = sizeof(long);
            break;
        }
        case 'e':
        {
            sz = sizeof(long long);
            break;
        }
        case 'f':
        {
            sz = sizeof(float);
            break;
        }
        case 'd':
        {
            sz = sizeof(double);
            break;
        }
        case 'p':
        {
            sz = sizeof(void*);
            break;
        }
        case 't':
        {
            sz = sizeof(void*);
            break;
        }
        case '2':
        {
            sz = sizeof(cinv_int16_t);
            break;
        }
        case '4':
        {
            sz = sizeof(cinv_int32_t);
            break;
        }
        case '8':
        {
            sz = sizeof(cinv_int64_t);
            break;
        }
        case 0:
        {
            // void
            break;
        }
        default:
        {
            assert(0);
        }
    }

    if (sz) ret_buf = malloc(sz);
    cinv_function_invoke(ctx, func, entrypoint, ret_buf, param_buf);
    
    for(; size > 0; size--)
    {
        free(param_buf[size - 1]);
    }
    free(param_buf);
    
    if (ret_buf)
    {
        switch(*ret_types_string)
        {
            case 'c':
            {
                tmp = kite_new_integer(thd, *(char*)ret_buf);
                break;
            }
            case 's':
            {
                tmp = kite_new_integer(thd, *(short*)ret_buf);
                break;
            }
            case 'i':
            {
                tmp = kite_new_integer(thd, *(int*)ret_buf);
                break;
            }
            case 'l':
            {
                tmp = kite_new_integer(thd, *(long*)ret_buf);
                break;
            }
            case 'e':
            {
                tmp = kite_new_integer(thd, (long)*(long long*)ret_buf);
                break;
            }
            case 'f':
            {
                tmp = kite_new_float(thd, *(float*)ret_buf);
                break;
            }
            case 'd':
            {
                tmp = kite_new_float(thd, *(double*)ret_buf);
                break;
            }
            case 'p':
            {
                tmp = kite_new_integer(thd, (long)*(void**)ret_buf);
                break;
            }
            case 't':
            {
                if (*(char**)ret_buf)
                    tmp = kite_new_string(thd, *(char**)ret_buf);
                else
                    tmp = kite_new_null(thd);
                break;
            }
            case '2':
            {
                tmp = kite_new_integer(thd, *(cinv_int16_t*)ret_buf);
                break;
            }
            case '4':
            {
                tmp = kite_new_integer(thd, *(cinv_int32_t*)ret_buf);
                break;
            }
            case '8':
            {
                tmp = kite_new_integer(thd, (long)*(cinv_int64_t*)ret_buf);
                break;
            }
            default:
            {
                assert(0);
            }
        }

        free(ret_buf);
        kite_vm_push(thd, tmp);
    }
    else
    {
        kite_vm_push(thd, kite_new_null(thd));
    }
}
#endif

KITE_MODULE_INITIALIZER(_cinvoke_function) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "_cinvoke_function");
    kite_set_docstring(newclass, "CInvokeFunction internal code.");

#ifdef BUILD_CINVOKE    
    /* Add methods. */
    kite_add_method(thread, newclass, "__construct__",
        kite_new_method_compiled_with_docs(thread, CInvokeFunction_construct, 
                                  "Class constructor.", 3,
                                  "ctx", "CInvokeContext instance",
                                  "ret", "Return parameter format",
                                  "param", "Input parameter format"));
    kite_add_method(thread, newclass, "destruct",
        kite_new_method_compiled_with_docs(thread, CInvokeFunction_destruct, 
                                  "Free associated memory and resources.", 0));
    kite_add_method(thread, newclass, "invoke",
        kite_new_method_compiled_with_docs(thread, CInvokeFunction_invoke, 
                                  "Retrieves function pointer for given function.", 2,
                                  "entry", "Function entry point.",
                                  "params", "List of parameters."));
#endif

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "_cinvoke_function", 1, NULL);
    kite_set_property(thread, parent, "_cinvoke_function", newclass);
}
